/*----------------------------------------------------------------------
    Line Cutter (by simple method)      lcut_s  ver1.0
        Written by  H.Goto, Nov. 1997
----------------------------------------------------------------------*/

/*--------------------------------------------------------------------
  Copyright (C) 1997  Hideaki Goto

        All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
  its documentation for any purpose is hereby granted without fee,
  provided that (i) the above copyright notice and this permission
  notice appear in all copies and in supporting documentation, (ii)
  the name of the author, Hideaki Goto, may not be used in any
  advertising or otherwise to promote the sale, use or other
  dealings in this software without prior written authorization
  from the author, (iii) this software may not be used for
  commercial products without prior written permission from the
  author, and (iv) the notice of the modification is specified in
  case of that the modified copies of this software are distributed.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
  THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
  SOFTWARE.
--------------------------------------------------------------------*/


/*
    Note:  You may not tune or change the following constants.
*/

#define		RectWidth	16		/* Width of PRs */
#define		RectParColumn	4		/* Rectangles par Column */

#define		C_nrect		8

#define		MAXSEGHEIGHT	332		/* > 40pt (at 600dpi) */

#define		C_clip		0.75		/* Const. for getting reference height */

#define		HCS		1.9		/* Const. for connection suppression */
#define		C_dist		0.5
#define		C_rheight	0.70

						/* Vertical Line Elimination */
#define		VLE_Width	2		/*   Threshold for width */
#define		VLE_Length	16		/*   Threshold for length */

						/* Separating character combination */
#define		C_sepchr_height	0.8		/*   Maximum height of candidates */
#define		C_sepchr_dist	0.4		/*   Maximum distance */
#define		C_sepchr_merged	1.2		/*   Maximum merged height */


/*
    Note: You may change the following constant arbitrarily.
*/

#define		MAXGROUPS	20000



#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>

#include	"utypes.h"

#include	"CRect.h"
#include	"objgrp.h"
#include	"ufilep.h"
#include	"siplib.h"

#define		MAXCONST	32767

typedef struct {
	int	pos,len;
} RUNLIST;


static int	sw_verbose = 0;
static int	sw_separate = 0;
static int	sw_NRpixels = 0;




/*----------------------------------------------
    Load PBM image file
----------------------------------------------*/

static int read_header(FILE *fp,int *width,int *height,int *linebytes){
    char  buf[80];
    int   ft = -1;
    int   maxgray;
    if ( NULL == fgets(buf,80,fp) )  return(-1);
    if ( 0 == strncmp("P4",buf,2) )  ft = 0;
    if ( 0 == strncmp("P5",buf,2) )  ft = 1;
    if ( -1 == ft )  return(-5);
    while (1){
        if ( NULL == fgets(buf,80,fp) )  return(-1);
        if ( *buf != '#' && *buf != '\n' )  break;
    }
    if ( 2 != sscanf(buf,"%d %d",&*width,&*height) )  return(-4);
        if ( ft == 0 )  *linebytes = (*width + 7) /8;
    if (0){
         printf("width=%d height=%d\n size=%d\n",*width,*height,*linebytes**height+1);
     }
    return(ft);
}


static SIPImage *load_image(char *fname,int *width,int *height){
	FILE	*fp;
	int	y,linebytes;
	uchar	mask,*p;
	SIPImage	*image;
	int	bufwidth;
	if ((fp = fopen(fname,"rb")) == NULL){
		fprintf(stderr,"Can't open file\n");
		return(0);
	}
	if (0 != read_header(fp,width,height,&linebytes)){
		fprintf(stderr,"File open error.\n");
		return(0);
	}

	mask = 0xff << (8 - (*width & 0x07));

	bufwidth = ((*width + RectWidth -1) / RectWidth) * RectWidth;
	if ( NULL == (image = sip_CreateImage(bufwidth,*height,1)) ){
		fputs("Low memory. (image buffer)\n",stderr);
		return(0);
	}

	for ( y=0 ; y < *height ; y++ ){
		p = (uchar *)sip_getimgptr(image,y);
		p[(bufwidth -1)/8] = 0;
		if ( linebytes != fread((char *)p,1,linebytes,fp) ){
			fprintf(stderr,"read error\n");
			return(0);
		}
		if ( mask )  p[*width /8] &= mask;
	}
	fclose(fp);
	return(image);
}




/*----------------------------------------------
    Calculate Distance between Two Lines
----------------------------------------------*/

static int get_linedistance(GRPLST *tlgroup,CRects *rectlist,int ncolumns,\
		int cwid,int grpid1,int grpid2,int *ulflag){
	int	*llist1,*llist2;
	int	*ulist1,*ulist2;
	int	*idlist;
	CRect	*rect;
	int	i,c,ac,d,d0;
	int	uc,lc;
	int	umin,lmin;
	if ( ! ncolumns )  return(-1);
	if ( 0 == (llist1 = new int[ncolumns * 4]) )  return(-1);
	llist2 = llist1 + ncolumns;
	ulist1 = llist2 + ncolumns;
	ulist2 = ulist1 + ncolumns;
	for ( i=0 ; i<ncolumns ; i++ ){
		llist1[i] = llist2[i] = MAXCONST;
		ulist1[i] = ulist2[i] = -1;
	}
	idlist = tlgroup->getidlist(grpid1);
	for ( i=0 ; i < tlgroup->getidcounts(grpid1) ; i++ ){
		rect = rectlist->getrect(idlist[i]);
		c = rect->x1 / cwid;
		if ( llist1[c] > rect->y1 )  llist1[c] = rect->y1;
		if ( ulist1[c] < rect->y2 )  ulist1[c] = rect->y2;
	}
	idlist = tlgroup->getidlist(grpid2);
	for ( i=0 ; i < tlgroup->getidcounts(grpid2) ; i++ ){
		rect = rectlist->getrect(idlist[i]);
		c = rect->x1 / cwid;
		if ( llist2[c] > rect->y1 )  llist2[c] = rect->y1;
		if ( ulist2[c] < rect->y2 )  ulist2[c] = rect->y2;
	}
	d = 0;		/* first, get reference distance */
	umin = lmin = MAXCONST;
	for ( i=uc=lc=0 ; i<ncolumns ; i++ ){
		if ( ulist1[i] == -1 )  continue;
		if ( ulist2[i] == -1 )  continue;
		if ( (llist1[i] + ulist1[i]) < (llist2[i] + ulist2[i]) ){
			if ( 0 > (d0 = llist2[i] - ulist1[i]) )  d0 = 0;
			if ( umin > d0 )  umin = d0;
			uc++;
		}
		else{	if ( 0 > (d0 = llist1[i] - ulist2[i]) )  d0 = 0;
			if ( lmin > d0 )  lmin = d0;
			lc++;
		}
		d += d0;
	}
	ac = uc + lc;
	if ( ac )  d /= ac;  else d = -1;

	if ( ac && ac < 3 ){	/* use minimum value instead of mean value,  */
				/*              if the overlap is very short */
		if ( umin < lmin ){  uc = 1;  lc = 0;  d = umin; }
		else              {  uc = 0;  lc = 1;  d = lmin; }
	}

	if ( ulflag ){
		if ( uc > lc )  *ulflag = 1;  else *ulflag = -1;
	}
	delete []llist1;
	return(d);
}




/*----------------------------------------------
    Sort line data and Create order table
----------------------------------------------*/

static int get_linepos(GRPLST *tlgroup,CRects *rect_pool,int gid){
	int	i,xpos,ypos;
	CRect	*rect;
	xpos = MAXCONST;
	ypos = 0;	/* default posision is y=0 */
	for ( i=0 ; i<tlgroup->getidcounts(gid) ; i++ ){
		rect = rect_pool->getrect( tlgroup->getidlist(gid)[i] );
		if ( xpos > rect->x1 ){
			xpos = rect->x1;
			ypos = (rect->y1 + rect->y2) /2;
		}
	}
	return(ypos);
}


static int get_linepos_v(GRPLST *tlgroup,CRects *rect_pool,int gid){
	int	i,xpos;
	CRect	*rect;
	xpos = MAXCONST;
	for ( i=0 ; i<tlgroup->getidcounts(gid) ; i++ ){
		rect = rect_pool->getrect( tlgroup->getidlist(gid)[i] );
		if ( xpos > rect->x1 ){
			xpos = rect->x1;
		}
	}
	return(xpos);
}


static void sort_horizontal(GRPLST *tlgroup,CRects *rect_pool,int *t_order, \
		int groups,int insertmode){
	int	i,j,k,m;
	for ( i=0 ; i<groups ; i++ ){		/* sort by horizontal posision */
		for ( j=i+1 ; j<groups ; j++ ){
			if (  get_linepos(tlgroup,rect_pool,t_order[i]) \
			    > get_linepos(tlgroup,rect_pool,t_order[j]) ){
				k = t_order[j];
				if ( insertmode ){
					for ( m=j-1 ; m>=i ; m-- )
						t_order[m+1] = t_order[m];
				}
				else{	t_order[j] = t_order[i];
				}
				t_order[i] = k;
			}
		}
	}
}


static void sort_vertical(GRPLST *tlgroup,CRects *rect_pool,int *t_order, \
		int groups,int insertmode){
	int	i,j,k,m;
	for ( i=0 ; i<groups ; i++ ){		/* sort by horizontal posision */
		for ( j=i+1 ; j<groups ; j++ ){
			if (  get_linepos_v(tlgroup,rect_pool,t_order[i]) \
			    > get_linepos_v(tlgroup,rect_pool,t_order[j]) ){
				k = t_order[j];
				if ( insertmode ){
					for ( m=j-1 ; m>=i ; m-- )
						t_order[m+1] = t_order[m];
				}
				else{	t_order[j] = t_order[i];
				}
				t_order[i] = k;
			}
		}
	}
}


static int sort_linedata(GRPLST *tlgroup,CRects *rect_pool,int ncolumns,\
		int cwid,int *t_order){
	int	i,j,k,m;
	int	gid1,gid2,groups,bgrp;
	int	ulflag;
	int	*ag;
	int	*rev_idx;
	for ( gid1=gid2=0 ; gid1 < tlgroup->groups ; gid1++ ){
		if ( tlgroup->getidcounts(gid1) == 0 )  continue;
		t_order[gid2++] = gid1;
	}
	groups = gid2;
	if ( groups == 0 )  return(0);
	if ( 0 == (ag = new int[groups * groups]) )  return(groups);
						/* exit doing nothing! */
	if ( 0 == (rev_idx = new int[tlgroup->groups]) ){
		delete []ag;
		return(0);			/* exit doing nothing! */
	}

	sort_horizontal(tlgroup,rect_pool,t_order,groups,0);
	sort_vertical(tlgroup,rect_pool,t_order,groups,1);

						/* create adjacency graph */
	for ( i=0 ; i<groups ; i++ ){
		gid1 = t_order[i];
		rev_idx[gid1] = i;
		for ( j=i+1 ; j<groups ; j++ ){
			k = get_linedistance(tlgroup,rect_pool,\
				ncolumns,cwid,gid1,t_order[j],&ulflag);
			if ( k < 0 ){
				ag[i * groups + j] = ag[j * groups + i] = 0;
				continue;
			}
			ag[j * groups + i] = -(ag[i * groups + j] = ulflag);
		}
	}

	for ( i=0 ; i<groups ; i++ ){		/* sort (insertion) */
		gid1 = t_order[i];
		m = groups - i;
		for ( j=i+1 ; j<groups ; j++ ){
			gid2 = t_order[j];
			ulflag = ag[ rev_idx[gid1] * groups + rev_idx[gid2] ];
			if ( ulflag == 0 )  continue;
			if ( ulflag < 0 ){
				gid1 = t_order[j];
				for ( k=j-1 ; k>=i ; k-- )  t_order[k+1] = t_order[k];
				t_order[i] = gid1;
				if ( m >= 0 ){	/* loop limitter */
					m--;
					j = i;
				}
			}
		}
	}

	delete []rev_idx;
	delete []ag;
	return(groups);
}




/*----------------------------------------------
    Export line data
----------------------------------------------*/

static int export_linedata(int width,int height,CRects *rect_pool,\
		GRPLST *tlgroup,int actgroups,int *t_order,char *fname){
	D4SAVE	D4SV;
	d4file	d4;
	d4pac	d4p;
	d4pac	d4p0 = { 0,0,0,0 };
	int	i,j,k,gid;
	int	retcode = 0;
	CRect	*rect;
	int	awidth,rwidth;
	d4.d4_attrb = 0;
	d4.d4_count = 0;
	d4.d4_llimit[0] = 0;	d4.d4_hlimit[0] = width -1;
	d4.d4_llimit[1] = 0;	d4.d4_hlimit[1] = height -1;
	d4.d4_llimit[2] = 0;	d4.d4_hlimit[2] = RectWidth;
	d4.d4_llimit[3] = 1;	d4.d4_hlimit[3] = height;
	if ( D4SV.filecreat(fname,&d4) )  return(-1);

	for ( k=0 ; k < actgroups && ! retcode ; k++ ){
		gid = t_order[k];
		for ( i=0 ; i < tlgroup->getidcounts(gid) && ! retcode ; i++ ){
			rect = rect_pool->getrect( tlgroup->getidlist(gid)[i] );
			d4p.d4_data[1] = rect->y1;
			d4p.d4_data[3] = rect->y2 - rect->y1 +1;
			awidth = rect->x2 - rect->x1 +1;
			for ( j=0 ; j<awidth ; j+=RectWidth ){
				rwidth = awidth - j;
				rwidth = rwidth < RectWidth ? rwidth : RectWidth;
				d4p.d4_data[0] = rect->x1 + j;
				d4p.d4_data[2] = rwidth;
				if ( D4SV.putdata(&d4p) )  retcode = -1;
			}
		}
		if ( D4SV.putdata(&d4p0) )  retcode = -1;
	}

	if ( D4SV.fileclose() )  return(-1);
	return(retcode);
}




/*----------------------------------------------
    Link segments
----------------------------------------------*/

static int link_segs(int width,int height,CRects *rect_pool,int ncolumn,\
		int *cltbl,GRPLST *tlgroup,int ref_height){
	int	i,j,k;
	int	cnum;
	int	dist,minrect;
	CRect	*rect1,*rect2;
	int	ct1,ct2;
	int	gid1,gid2;
	int	h1,h2;

	for ( i=0 ; i<cltbl[1] ; i++ ){
		rect1 = rect_pool->getrect(i);
		if ( 0 > (gid1 = tlgroup->newgrp(i)) )  return(-1);
		rect1->attr = gid1;
	}

	for ( cnum=0 ; cnum < ncolumn -1 ; cnum++ ){

		/* left -> right connection */
		for ( i=cltbl[cnum] ; i<cltbl[cnum+1] ; i++ ){
			dist = height;  minrect = -1;
			rect1 = rect_pool->getrect(i);
			ct1 = (rect1->y1 + rect1->y2) /2;
			gid1 = rect1->attr;
			if ( gid1 < 0 ){
				if ( 0 > (gid1 = tlgroup->newgrp(i)) ){
					return(-1);
				}
				rect1->attr = gid1;
			}
			h1 = (int)(rect1->y2 - rect1->y1 +1);
			if ( h1 > HCS * ref_height )  continue;
			/* find nearest segment */
			for ( j=cltbl[cnum+1] ; j<cltbl[cnum+2] ; j++ ){
				rect2 = rect_pool->getrect(j);
				h2 = (int)(rect2->y2 - rect2->y1 +1);
				if ( h2 > HCS * ref_height )  continue;
				ct2 = (rect2->y1 + rect2->y2) /2;
				if ( dist > abs(ct1-ct2) ){
					dist = abs(ct1-ct2);
					minrect = j;
				}
			}
			if ( minrect == -1 )  continue;
			rect2 = rect_pool->getrect(minrect);
			if ( rect1->y1 > rect2->y2 )  continue;
			if ( rect1->y2 < rect2->y1 )  continue;
			gid2 = rect2->attr;
			if ( gid2 < 0 ){
				if ( 0 > tlgroup->addgrp(gid1,minrect) )  return(-2);
				rect2->attr = gid1;
			}
			else if ( gid1 != gid2 ){
				for ( k=0 ; k<tlgroup->getidcounts(gid2) ; k++ ){
					rect_pool->getrect( tlgroup->getidlist(gid2)[k] )->attr = gid1;
				}
				if ( 0 > tlgroup->mergegrps(gid1,gid2) )  return(-3);
				tlgroup->deletegrp(gid2);
			}
		}

		/* left <- right connection */
		for ( i=cltbl[cnum+1] ; i<cltbl[cnum+2] ; i++ ){
			dist = height;  minrect = -1;
			rect1 = rect_pool->getrect(i);
			ct1 = (rect1->y1 + rect1->y2) /2;
			gid1 = rect1->attr;
			if ( gid1 < 0 ){
				if ( 0 > (gid1 = tlgroup->newgrp(i)) ){
					return(-1);
				}
				rect1->attr = gid1;
			}
			h1 = (int)(rect1->y2 - rect1->y1 +1);
			if ( h1 > HCS * ref_height )  continue;
			/* find nearest segment */
			for ( j=cltbl[cnum] ; j<cltbl[cnum+1] ; j++ ){
				rect2 = rect_pool->getrect(j);
				h2 = (int)(rect2->y2 - rect2->y1 +1);
				if ( h2 > HCS * ref_height )  continue;
				ct2 = (rect2->y1 + rect2->y2) /2;
				if ( dist > abs(ct1-ct2) ){
					dist = abs(ct1-ct2);
					minrect = j;
				}
			}
			if ( minrect == -1 )  continue;
			rect2 = rect_pool->getrect(minrect);
			if ( rect1->y1 > rect2->y2 )  continue;
			if ( rect1->y2 < rect2->y1 )  continue;
			gid2 = rect2->attr;
			if ( gid2 < 0 ){
				if ( 0 > tlgroup->addgrp(gid1,minrect) )  return(-2);
				rect2->attr = gid1;
			}
			else if ( gid1 != gid2 ){
				for ( k=0 ; k<tlgroup->getidcounts(gid2) ; k++ ){
					rect_pool->getrect( tlgroup->getidlist(gid2)[k] )->attr = gid1;
				}
				if ( 0 > tlgroup->mergegrps(gid1,gid2) )  return(-3);
				tlgroup->deletegrp(gid2);
			}
		}

	}
	return(0);
}




/*----------------------------------------------
    Calculate Average Line Height
----------------------------------------------*/

static int get_avrlineheight(GRPLST *tlgroup,CRects *rectlist,int grpid,int clip){
	int	*idlst,nrect,i,n,hsum0,hsum,h;
	CRect	*rect;
	idlst = tlgroup->getidlist(grpid);
	nrect = tlgroup->getidcounts(grpid);
	hsum0 = 0;
	for ( i=0 ; i<nrect ; i++ ){
		rect = rectlist->getrect(idlst[i]);
		h = (int)(rect->y2 - rect->y1 +1);
		hsum0 += h;
	}
	if ( nrect )  hsum0 /= nrect;

	if ( ! clip )  return(hsum0);

	hsum = 0;
	for ( i=n=0 ; i<nrect ; i++ ){
		rect = rectlist->getrect(idlst[i]);
		h = (int)(rect->y2 - rect->y1 +1);
		if ( h > (C_clip * hsum0) ){
			hsum += h;
			n++;
		}
	}
	if ( n == 0 )  return(hsum0);
	return(hsum / n);
}




/*----------------------------------------------
    Calculate Merged Line Height
----------------------------------------------*/

static int get_mergedheight(GRPLST *tlgroup,CRects *rectlist,int ncolumns,\
		int cwid,int grpid1,int grpid2){
	int	*llist1,*llist2;
	int	*ulist1,*ulist2;
	int	*idlist;
	CRect	*rect;
	int	i,c,ac,d;
	int	l,u;
	if ( ! ncolumns )  return(-1);
	if ( 0 == (llist1 = new int[ncolumns * 4]) )  return(-1);
	llist2 = llist1 + ncolumns;
	ulist1 = llist2 + ncolumns;
	ulist2 = ulist1 + ncolumns;
	for ( i=0 ; i<ncolumns ; i++ ){
		llist1[i] = llist2[i] = MAXCONST;
		ulist1[i] = ulist2[i] = -1;
	}
	idlist = tlgroup->getidlist(grpid1);
	for ( i=0 ; i < tlgroup->getidcounts(grpid1) ; i++ ){
		rect = rectlist->getrect(idlist[i]);
		c = rect->x1 / cwid;
		if ( llist1[c] > rect->y1 )  llist1[c] = rect->y1;
		if ( ulist1[c] < rect->y2 )  ulist1[c] = rect->y2;
	}
	idlist = tlgroup->getidlist(grpid2);
	for ( i=0 ; i < tlgroup->getidcounts(grpid2) ; i++ ){
		rect = rectlist->getrect(idlist[i]);
		c = rect->x1 / cwid;
		if ( llist2[c] > rect->y1 )  llist2[c] = rect->y1;
		if ( ulist2[c] < rect->y2 )  ulist2[c] = rect->y2;
	}
	d = 0;
	for ( i=ac=0 ; i<ncolumns ; i++ ){
		if ( ulist1[i] == -1 )  continue;
		if ( ulist2[i] == -1 )  continue;
		l = llist1[i];  if ( l > llist2[i] )  l = llist2[i];
		u = ulist1[i];  if ( u < ulist2[i] )  u = ulist2[i];
		d += u - l;
		ac++;
	}
	if ( ac )  d /= ac;  else d = -1;
	delete []llist1;
	return(d);
}



/*------------------------------------------------------
    Combine elements of separating character
------------------------------------------------------*/

static int combine_sepchr(GRPLST *tlgroup,CRects *rectlist,int ncolumns,\
		int cwid,int ref_height){
	int	i,j,k,gid1,gid2;
	int	*tgid;
	int	ntext;
	int	dist;
	if ( ! tlgroup->groups )  return(0);
	if ( 0 == (tgid = new int[tlgroup->groups]) )  return(-1);
	ntext = 0;
	for ( gid1=0 ; gid1 < tlgroup->groups ; gid1++ ){
		if ( tlgroup->getidcounts(gid1) == 0 )  continue;
		if ( tlgroup->getidcounts(gid1) > 2 )  continue;
		if ( get_avrlineheight(tlgroup,rectlist,gid1,1) \
			> C_sepchr_height * ref_height )  continue;
		tgid[ntext++] = gid1;
	}
	for ( i=0 ; i<ntext ; i++ ){
		gid1 = tgid[i];
		for ( j=i+1 ; j<ntext ; j++ ){
			gid2 = tgid[j];
			dist = get_linedistance(tlgroup,rectlist,ncolumns,\
						cwid,gid1,gid2,NULL);
			if ( dist < 0 )  continue;
			if ( dist > C_sepchr_dist * ref_height )  continue;
			if ( get_mergedheight(tlgroup,rectlist,ncolumns,\
			  cwid,gid1,gid2) > C_sepchr_merged * ref_height )  continue;
			for ( k=0 ; k<tlgroup->getidcounts(gid1) ; k++ ){
				rectlist->getrect( tlgroup->getidlist(gid1)[k] )->attr = gid2;
			}
			if ( 0 > tlgroup->mergegrps(gid2,gid1) )  return(-1);
			tlgroup->deletegrp(gid1);
		}
	}
	delete []tgid;
	return(0);
}




/*------------------------------------------------------
    Combine underlines/overlines/ruby characters
------------------------------------------------------*/

static int combine(GRPLST *tlgroup,CRects *rectlist,int ncolumns,\
		int cwid,int ref_height,int ulflag){
	int	i,j,k,gid1,gid2,mingid;
	int	*avrheight;
	int	*rgid,*tgid;
	int	nruby,ntext;
	int	dist,mindist,mindist_w,w;
	int	ul;
	if ( ! tlgroup->groups )  return(0);
	if ( 0 == (avrheight = new int[tlgroup->groups * 3]) )  return(-1);
	rgid = avrheight + tlgroup->groups;
	tgid = rgid + tlgroup->groups;
	nruby = ntext = 0;
	for ( gid1=0 ; gid1 < tlgroup->groups ; gid1++ ){
		if ( tlgroup->getidcounts(gid1) == 0 )  continue;
		avrheight[gid1] = get_avrlineheight(tlgroup,rectlist,gid1,1);
		if ( avrheight[gid1] >= C_rheight * ref_height ){
			tgid[ntext++] = gid1;
		}
		else{	rgid[nruby++] = gid1;
		}
	}
	for ( i=0 ; i<nruby ; i++ ){
		gid1 = rgid[i];
		mindist = mindist_w = MAXCONST;
		mingid = -1;
		for ( j=0 ; j<ntext ; j++ ){
			gid2 = tgid[j];
			dist = get_linedistance(tlgroup,rectlist,ncolumns,\
						cwid,gid1,gid2,&ul);
			if ( ulflag == 0 )        w = 1;
			else if ( ul == ulflag )  w = 2;
			else                      w = 3;
			if ( dist >= 0 && mindist_w > w * dist ){
				mingid = gid2;
				mindist = dist;
				mindist_w = w * dist;
			}
		}
		if ( mingid != -1 ){
			if ( mindist <= C_dist * ref_height ){
				for ( k=0 ; k<tlgroup->getidcounts(gid1) ; k++ ){
					rectlist->getrect( tlgroup->getidlist(gid1)[k] )->attr = mingid;
				}
				if ( 0 > tlgroup->mergegrps(mingid,gid1) )  return(-1);
				tlgroup->deletegrp(gid1);
			}
		}
	}
	delete []avrheight;
	return(0);
}




/*----------------------------------------------
    Get reference height of rectangle
----------------------------------------------*/

static long get_ref_height(CRects *rectlist,int avr_height){
	int	rid;
	CRect	*rect;
	long	ref_height = 0;
	long	ref_counts = 0;
	int	height;
	for ( rid=0 ; rid < rectlist->counts ; rid++ ){
		rect = rectlist->getrect(rid);
		height = rect->y2 - rect->y1 +1;
		if ( height > (C_clip * avr_height) ){
			ref_height += (long)height;
			ref_counts++;
		}
	}
	if ( ref_counts )  ref_height /= ref_counts;
	return(ref_height);
}




/*----------------------------------------------
    Create Local Rectangles
----------------------------------------------*/

			/* Note: This is an easy method; never aimed for perfect. */
static int filter_proj(long *proj,int len,int th,int th_len){
	int	i,j,l;
	for ( i=l=0 ; i<len ; ){
		for ( ; i<len && (proj[i] == 0 || proj[i] > th) ; i++ ) ;
		for ( l = 0; i<len && proj[i] != 0 && proj[i] <= th ; i++ ) l++;
		if ( l >= th_len ){
			for ( j=i-l ; j < i ; j++ )  proj[j] = 0;
		}
		l = 0;
	}
	if ( l >= th_len ){
		for ( j=i-l ; j < i ; j++ )  proj[j] = 0;
	}
	return(0);
}


static int create_rects(SIPImage *image,int x,int awidth, \
		RUNLIST *runlist,int minpix){
	int	i,y,l,pixsum,height;
	long	*rden;
	SIPRectangle	cbox;
	cbox.x = (short)x;
	cbox.y = 0;
	cbox.width = (ushort)awidth;
	cbox.height = (ushort)image->height;
	height = (int)image->height;
	if ( 0 == (rden = new long[height]) )  return(-1);
	sip_projprofile_area(image,rden,0,&cbox);
	filter_proj(rden,height,VLE_Width,VLE_Length);
	for ( y=i=0 ; y<height ; ){
		for ( l=0 ; rden[y] <= 0 && y<height ; y++ ) ;
		if ( y >= height )  break;
		runlist[i].pos = y;
		for ( l=pixsum=0 ; rden[y] > 0 && y<height ; y++ ){
			l++;
			pixsum += rden[y];
		}
		runlist[i].len = l;
		if ( pixsum >= minpix )  i++;
	}
	runlist[i].pos = -1;
	delete []rden;
	return(0);
}




/*----------------------------------------------
    Print usage
----------------------------------------------*/

static void usage(){
	fputs("lcut_s v1.0, Copyright (C) 1997 Hideaki Goto \n",stderr);
	fputs("Usage: lcut_s [-options] in_file out_file \n",stderr);
	fputs("         -nr:       apply noise reduction \n",stderr);
	fputs("         -global:   use global projection profile \n",stderr);
	fputs("         -separate: suppress combining lines and ruby characters \n",stderr);
	fputs("         -verbose:  enable printing processing information \n",stderr);
	fputs("         in_file:   image data in PBM(raw,binary) format \n",stderr);
	fputs("         out_file:  text line data in D4-format \n",stderr);
}




/*----------------------------------------------
    Main routine
----------------------------------------------*/

int main(int ac,char *av[]){
	char	*infile  = NULL;
	char	*outfile = NULL;
	int	i,j,k,x,y;
	int	width,height,awidth;
	long	avr_height;
	long	ref_height;
	SIPImage	*image;
	int	rect_par_column = RectParColumn;
	int	cwid;
	int	cnum;
	int	*cltbl;
	int	rects;
	RUNLIST	*runlist;
	int	crn;
	CRects	rect_pool;
	GRPLST	tlgroup;
	int	actgroups;
	int	*t_order;

	for ( k=1, i=1 ; i<ac && k ; i++ ){
		if ( 0 == strcmp("-nr",av[i]) ){  sw_NRpixels = 4;  continue; }
		if ( 0 == strcmp("-global",av[i]) ){  rect_par_column = -1;  continue; }
		if ( 0 == strcmp("-separate",av[i]) ){  sw_separate = -1;  continue; }
		if ( 0 == strcmp("-verbose",av[i]) ){  sw_verbose = -1;  continue; }
		if ( 0 == strcmp("-v",av[i]) ){  sw_verbose = -1;  continue; }
		if ( av[i][0] == '-' ){  k=0;  continue; }
		if ( NULL == infile ){  infile = av[i];  continue; }
		if ( NULL == outfile ){  outfile = av[i];  continue; }
	}
	if ( k == 0 || outfile == NULL ){  usage();  return(0); }


/* ---- Read Image ---- */

	if ( 0 == (image = load_image(infile,&width,&height)) )  return(1);
	if ( sw_verbose )
		printf("Image size = %d x %d\n",width,height);


/* ---- Create Rectangles ---- */

	if ( 0 == (runlist = new RUNLIST[height +1]) ){
		fputs("Low memory.\n",stderr);
		exit(1);
	}

	if ( rect_par_column == -1 )
		rect_par_column = (width + RectWidth -1)/RectWidth;

	while(1){
		cwid = rect_par_column * RectWidth;
		cnum = (width + cwid -1) / cwid;
		if ( 0 == (cltbl = new int[cnum +1]) ){
			fputs("Low memory.\n",stderr);
			exit(1);
		}

		cnum = rects = 0;
		avr_height = 0;
		rect_pool.clear();
		for ( x=0 ; x<width ; x+=cwid ){
			cltbl[cnum] = rects;
			awidth = width - x;
			awidth = awidth < cwid ? awidth : cwid;
			create_rects(image,x,awidth,runlist,sw_NRpixels);
			for ( i=0 ; runlist[i].pos >= 0 ; i++ ){
				if ( runlist[i].len > MAXSEGHEIGHT )  continue;
				crn = rect_pool.create(x,runlist[i].pos, \
					x + (awidth -1), runlist[i].pos + runlist[i].len -1);
				avr_height += (long)runlist[i].len;
				if ( 0 > crn ){
					fputs("Low memory. (rectangles)\n",stderr);
					exit(1);
				}
				rect_pool.setattr(crn,-1);
					/* attribute = -1: not yet linked */
				if ( sw_verbose ){
					printf("%3d : %03d %03d %03d\n", \
						rects,x,runlist[i].pos,runlist[i].len);
				}
				rects++;
			}
			cnum++;
		}
		cltbl[cnum] = rects;
		if ( rects )  avr_height /= rects;
		ref_height = get_ref_height(&rect_pool,avr_height);
		if ( rect_par_column == RectParColumn ){
			if ( (ref_height / C_nrect) > RectParColumn ){
				rect_par_column = ref_height / C_nrect;
				delete []cltbl;
				continue;
			}
		}
		break;
	}

	if ( sw_verbose )
		printf("rect_par_column = %d\n",rect_par_column);

/* ---- Link segments ---- */

	tlgroup.clearall();
	tlgroup.alloc(MAXGROUPS);
	link_segs(width,height,&rect_pool,cnum,cltbl,&tlgroup,ref_height);


/* ---- Combine underlines/overlines/ruby characters ---- */

	if ( ! sw_separate ){
		combine(&tlgroup,&rect_pool,cnum,cwid,ref_height,1);
		combine_sepchr(&tlgroup,&rect_pool,cnum,cwid,ref_height);
	}


/* ---- Export line data ---- */

	actgroups = 0;
	if ( tlgroup.groups > 0 ){
		if ( 0 == (t_order = new int[tlgroup.groups]) ){
			fputs("Low memory. (text line sorting)\n",stderr);
			exit(1);
		}
		actgroups = sort_linedata(&tlgroup,&rect_pool,cnum,cwid,t_order);

		if ( sw_verbose ){
			for ( k=0 ; k<actgroups ; k++ ){
				i = t_order[k];
				printf("Group=%d (%d):\n",i,tlgroup.getidcounts(i));
				for ( j=0 ; j<tlgroup.getidcounts(i) ; j++ ){
					printf(" %d",tlgroup.getidlist(i)[j]);
				}
				printf("\n");
			}
		}
	}

	if ( export_linedata(width,height,&rect_pool,&tlgroup,actgroups,t_order,outfile) ){
		fprintf(stderr,"Write to %s failed.\n",outfile);
		exit(3);
	}

	if ( tlgroup.groups > 0 )  delete []t_order;


	delete []cltbl;
	return(0);
}


