.TH lcut_s 1 "Nov. 1997"
.SH NAME
lcut_s \- extract text lines from a text block image
.SH SYNOPSIS
.B lcut_s
.BR
[\-options]
.BR
.I in_file
.BR
.I out_file

.SH DESCRIPTION
lcut_s is the implementation of the text line extraction by 
a relatively simple method.
lcut_s reads a text block image from
.I in_file,
separate text lines in 
the image, and writes the outline data on the specified output file
.I out_file.

The strong point of this program is that the existence of
ruby characters in Japanese document, underlines and overlines
are taken into account.
Ruby characters here are the printed Japanese characters alongside
Chinese characters.
These characters and lines are the attendant elements on
main text lines, and they should be analyzed and recognized together.
lcut_s can combine these elements with main text lines.

.PP
The input image is expected to be the text block image;
the image of text lines which are previously cut out of the page image.
lcut_s assumes that the text lines in the image are all horizontally written,
and that the characters are put on upright or rotated counterclockwise
by 90degs as shown below.

.nf
+\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-> +x
|
|   ##    #####
|  #  #   #    #
| #    #  #####
| ######  #    #
| #    #  #    #
| #    #  ####
|
v +y
.fi

.nf
+\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-\-> +x
|
|      ##########     ####   ####
|   ###    #         #    # #    #
| ##       #        #      #      #
|   ###    #        #      #      #
|      ##########   ###############
|
v +y
.fi

The input image should be roughly skew corrected before processing.
However,
it is not necessary to apply any precise skew correction before, 
because lcut_s is tolerant of slight skew of text lines.
This advantageous property comes from the algorithm of 
text line extraction based on local projection profile.
Note that the limit of the skew angle depend on the input image.
A few degrees (\-3degs to +3degs) might be a good guess.

.PP
The file format of 
.I in_file
is raw, binary pbm(5).
Though lcut_s can handle images in various resolution,
200dpi to 600dpi is better.
lcut_s cannot extract text lines
with very large (>40pt) characters.
The input image should not contain text lines by 
various font size, because
the font size of main text lines are expected to be uniform.

The file format of the output file is tloutline(4).
You can obtain each image of text line by
strextr(1) command with \-raw option.
The typical operation is as shown below.
(The image files are created in ./cuttmp subdirectory.)

.RS 4
.nf
example% rm \-fr ./cuttmp ; mkdir ./cuttmp
example% lcut_s in_file out_file
example% strextr \-raw \-image in_file out_file
.fi
.RE


.SH OPTIONS
.TP
.B \-nr
Reduce tiny, noisy points in the input image.
.TP
.B \-global
Use global projection profile to separate text lines.
This option speed up the processing.
However, the process gets very sensitive to
the skew of text lines if this option is specified.
This is the week point of global method.
.TP
.B \-separate
Suppress combining underlines, overlines and ruby characters
with main text lines.
Note that lcut_s never try to separate elements which are touching each other.
.TP
.B \-verbose 
Enable printing processing information.


.SH EXIT STATUS
The following exit values are returned:
.PD 0
.TP 10
.B 0
Successful completion.
.TP
.B > 0
An error occurred.
.PD

.SH SEE ALSO
strextr(1), tloutline(4), pbm(5)

.SH NOTES
lcut_s was designed to process text block images.
In general, lcut_s does not work well for page images.
.PP
lcut_s may fail to separate some main text lines,
if inter\-line spacing is extremely narrow in the text block 
being processed.
.PP
Ruby characters, underlines and overlines often touch
the characters in main text line.  It is not easy to
separate these elements correctly.


.SH AUTHOR
Copyright (C) 1997 Hideaki Goto
.\" Permission to use, copy, modify, and distribute this software and
.\" its documentation for any purpose is hereby granted without fee,
.\" provided that (i) the above copyright notice and this permission
.\" notice appear in all copies and in supporting documentation, (ii)
.\" the name of the author, Hideaki Goto, may not be used in any
.\" advertising or otherwise to promote the sale, use or other
.\" dealings in this software without prior written authorization
.\" from the author, (iii) this software may not be used for
.\" commercial products without prior written permission from the
.\" author, and (iv) the notice of the modification is specified in
.\" case of that the modified copies of this software are distributed.
.\"
.\" THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND.
.\" THE AUTHOR WILL NOT BE RESPONSIBLE FOR ANY DAMAGE CAUSED BY THIS
.\" SOFTWARE.

E\-Mail: hgot@ecip.tohoku.ac.jp, hgot@aso.ecei.tohoku.ac.jp
